﻿using System;
using System.Collections.Generic;
using System.Reflection;
using System.Runtime.CompilerServices;
using System.Threading;

namespace ConsoleApp1
{
    class Program
    {
        List<Thread> _threads = new List<Thread>();
        static volatile bool _safeStackTrace;

        static void Main(string[] args)
        {
            GC.RegisterForFullGCNotification(10, 10);

            _safeStackTrace = true;

            Thread thWaitForFullGC = new Thread(GCNotifyProc);
            thWaitForFullGC.IsBackground = true;
            thWaitForFullGC.Start();

            Program pg = new Program();
            pg.Start();
            Console.ReadLine();
        }

        static private void GCNotifyProc()
        {
            while (true)
            {
                var gcApproachStatus = GC.WaitForFullGCApproach(); 
                if (gcApproachStatus == GCNotificationStatus.Succeeded)
                {
                    _safeStackTrace = false;
                    GC.Collect();
                }
                else
                {
                    return;
                }

                var gcCompleteStatus = GC.WaitForFullGCComplete();
                if (gcCompleteStatus == GCNotificationStatus.Succeeded)
                {
                    _safeStackTrace = true;
                }
                else
                {
                    return;
                }
            }
        }

        int _count = 0;
        int _gcCount = 0;

        private void Start()
        {
            for (int i = 0; i < 4; i++)
            {
                Thread t = new Thread(threadFunc);
                _threads.Add(t);
                t.IsBackground = true;
                t.Name = i.ToString();
                t.Start();
            }

            Thread.Sleep(5000);
            Random rd = new Random(Environment.TickCount);

            while (true)
            {
                _count++;
                int idx = rd.Next(0, _threads.Count - 1);
                Thread t = _threads[idx];

                {
                    GetCallStack(t);
                }

                _gcCount = GC.CollectionCount(2);

                if (_count % 100 == 0)
                {
                    Console.WriteLine(_count + ": " + _gcCount);
                }
            }
        }

        [System.Security.SecuritySafeCritical]
        private static string GetCallStack(Thread t)
        {
            string result = "";
            try
            {

                System.Diagnostics.StackTrace trace = null;
                t.Suspend();

                RuntimeHelpers.PrepareConstrainedRegions();
                try
                {
                }
                finally
                {
                    if (_safeStackTrace == true)
                    {
                        trace = new System.Diagnostics.StackTrace(t, false);
                        result = trace.ToString();
                    }

                    try
                    {
                        t.Resume();
                    }
                    catch { }
                }

            } catch { }

            return result;
        }


        private static void threadFunc()
        {
            List<byte[]> _bufs = new List<byte[]>();

            while (true)
            {
                byte[] buf1 = new byte[4096 * 512];
                lock (_bufs)
                {
                    _bufs.Add(buf1);

                    if (_bufs.Count >= 2)
                    {
                        _bufs.Clear();
                    }
                }

                Thread.Sleep(10);
            }
        }
    }
}
